from diffgram.core.core import Project
import json
import os
import requests
import io
import math
import uuid


class File3D:
    point_list: list
    client: Project

    def __init__(self, client, name, point_list = []):
        self.point_list = []
        self.client = client
        self.original_filename = name
        self.point_list = point_list

    def add_point(self,
                  x: float,
                  y: float,
                  z: float,
                  intensity = None,
                  device_id = None,
                  timestamp = None,
                  is_ground = False):

        if intensity is not None:
            if intensity > 1.0 or intensity < 0:
                raise Exception('Intensity point must be between 0 and 1. Value is: {}'.format(intensity))

        self.point_list.append({
            'x': x,
            'y': y,
            'z': z,
            'intensity': intensity,
            'device_id': device_id,
            'timestamp': timestamp,
            'is_ground': is_ground,
        })
        return self.point_list

    def upload(self, dataset_name = None, chunk_size = 5000000):
        """
            Builds a JSON file from current point list and uploads it to
            Diffgram.
        :param dataset_name:
        :param chunk_size: Size of each chunk of the JSON file. Default is 5MB
        :return:
        """

        if len(self.point_list) == 0:
            print('At least 1 point should be provided in the point list. Please add point using add_point()')
            return False
        file_data = {
            'point_list': self.point_list
        }
        json_data = json.dumps(file_data)

        with open('data.json', 'w') as outfile:
            json.dump(json_data, outfile)
        endpoint = "/api/walrus/project/{}/upload/large".format(
            self.client.project_string_id
        )
        chunk_size = 5000000  # 5 MB chunks
        dataset_id = self.client.default_directory.id
        if dataset_name is not None:
            dataset_id = self.client.directory.get(dataset_name).id

        with io.StringIO(json_data) as s:
            pos = s.tell()
            s.seek(0, os.SEEK_END)
            file_size = s.tell()
            s.seek(pos)
            num_chunks = int(math.ceil(file_size / chunk_size))
            last_chunk_size = -1
            if file_size % chunk_size != 0:
                last_chunk_size = file_size % chunk_size

            uid_upload = str(uuid.uuid4())
            for i in range(0, num_chunks):

                payload = {
                    'dzuuid': uid_upload,
                    'original_filename': self.original_filename,
                    'dzchunkindex': i,
                    'dztotalfilesize': file_size,
                    'dzchunksize': chunk_size,
                    'dztotalchunkcount': num_chunks,
                    'dzchunkbyteoffset': i * chunk_size,
                    'directory_id': dataset_id,
                    'source': 'from_sensor_fusion_json',
                }
                # Adjust final chunk size
                if i == (num_chunks - 1) and last_chunk_size != -1:
                    # Read last chunk size here...
                    payload['dzchunksize'] = last_chunk_size

                # Read file Chunk
                s.seek(payload['dzchunkbyteoffset'])
                file_chunk = s.read(payload['dzchunksize'])
                files = {'file': ('{}_sf.json'.format(self.original_filename), file_chunk)}
                # Make request to server
                url = self.client.host + endpoint
                response = self.client.session.post(url,
                                                    data = payload,
                                                    files = files)

                self.client.handle_errors(response)

            return True



import warnings


def label_new(self,
              label: dict,
              schema_id: int = None,
              allow_duplicates: bool = False,
              print_success: bool = True):
    """

    Arguments
        self,
        label_list, a list of label strings
        ignore_duplicates, bool
        print_success, bool

    Expects

    Returns

    """
    if schema_id is None:
        schema = self.get_default_label_schema()
        if schema is not None:
            schema_id = schema.get('id')

    # Check if already exists
    name = label.get('name', None)
    if not name:
        raise Exception("Please provide a key of name with a value of label")
    label['schema_id'] = schema_id
    if allow_duplicates is False:

        label_file_id = self.name_to_file_id.get(name, None)

        if label_file_id:
            warnings.warn("\n\n '" + name + "' label already exists and was skipped." + \
                          "\n Set allow_duplicates = True to bypass this check. \n")
            return

    endpoint = "/api/v1/project/" + self.project_string_id + \
               "/label/new"

    response = self.session.post(self.host + endpoint,
                                 json = label)

    data = response.json()
    self.get_label_file_dict()
    if data["log"]["success"] == True:
        if print_success is True:
            print("New label success")
    else:
        raise Exception(data["log"]["error"])



import requests

from diffgram import __version__

from diffgram.file.view import get_label_file_dict
from diffgram.convert.convert import convert_label
from diffgram.label.label_new import label_new

from diffgram.core.directory import Directory
from diffgram.job.job import Job
from diffgram.job.guide import Guide
from diffgram.brain.brain import Brain
from diffgram.file.file_constructor import FileConstructor
from diffgram.file.file import File
from diffgram.brain.train import Train
from diffgram.export.export import Export
from diffgram.task.task import Task
from requests.auth import HTTPBasicAuth


class Project():

    def __init__(
            self,
            project_string_id,
            client_id = None,
            client_secret = None,
            debug = False,
            staging = False,
            host = None,
            init_default_directory = True,
            refresh_local_label_dict = True

    ):

        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(pool_connections = 30, pool_maxsize = 30)
        self.session.mount('http://', adapter)
        self.session.mount('https://', adapter)
        self.project_string_id = None

        self.debug = debug
        self.staging = staging
        if host is None:
            if self.debug is True:
                self.host = "http://127.0.0.1:8085"
                print("Debug", __version__)
            elif self.staging is True:
                self.host = "https://20200110t142358-dot-walrus-dot-diffgram-001.appspot.com/"
            else:
                self.host = "https://diffgram.com"
        else:
            self.host = host

        self.auth(
            project_string_id = project_string_id,
            client_id = client_id,
            client_secret = client_secret)

        self.file = FileConstructor(self)
        #self.train = Train(self)
        self.job = Job(self)
        self.guide = Guide(self)
        self.directory = Directory(self, 
                                   init_file_ids = False,
                                   validate_ids = False)
        self.export = Export(self)
        self.task = Task(client = self)

        self.directory_id = None
        self.name_to_file_id = None


        if init_default_directory is True:
            self.set_default_directory(directory = self.directory)
            print("Default directory set:", self.directory_id)

        if refresh_local_label_dict is True:
            self.get_label_file_dict()

        self.client_id = client_id
        self.client_secret = client_secret

        self.label_schema_list = self.get_label_schema_list()

        self.directory_list = None


    def get_member_list(self):
        url = '/api/project/{}/view'.format(self.project_string_id)
        response = self.session.get(url = self.host + url)
        self.handle_errors(response)
        data = response.json()
        return data['project']['member_list']

    def get_label_schema_by_id(self, id):
        if self.label_schema_list is None or len(self.label_schema_list) == 0:
            self.label_schema_list = self.get_label_schema_list()
        for s in self.label_schema_list:
            if s['id'] == id:
                return s

    def get_label_schema_by_name(self, name):
        if self.label_schema_list is None or len(self.label_schema_list) == 0:
            self.label_schema_list = self.get_label_schema_list()
        for s in self.label_schema_list:
            if s['name'] == name:
                return s

    def get_default_label_schema(self):
        if self.label_schema_list is None or len(self.label_schema_list) == 0:
            self.label_schema_list = self.get_label_schema_list()

        return self.label_schema_list[0]

    def get_label_list(self, schema_id = None):
        url = f'/api/project/{self.project_string_id}/labels'
        if schema_id is None:
            schema = self.get_default_label_schema()
            if schema is not None:
                schema_id = schema.get('id')

        params = {'schema_id': schema_id}
        response = self.session.get(url = self.host + url, params=params)
        self.handle_errors(response)
        data = response.json()
        return data.get('labels_out')

    def get_label_schema_list(self):
        url = f'/api/v1/project/{self.project_string_id}/labels-schema'
        response = self.session.get(url = self.host + url)
        self.handle_errors(response)
        data = response.json()
        return data

    def get_http_auth(self):
        return HTTPBasicAuth(self.client_id, self.client_secret)

    def get_label(
            self,
            name = None,
            schema_id = None,
            name_list = None):
        """
        name, str
        name_list, list, optional

        Name must be an exact match to label name.

        If a name_list is provided it will construct a list of
        objects that match that name.

        Returns
            None if not found.
            File object of type Label if found.
            List of File objects if a proj is provided.
        """
        if self.name_to_file_id is None:
            self.get_label_file_dict()

        if name_list:
            out = []
            for name in name_list:
                out.append(self.get_label(name))
            return out

        id = self.name_to_file_id.get(name)

        if id is None:
            return None

        file = File(id = id)
        return file

    def get_model(
            self,
            name = None,
            local = False):

        brain = Brain(
            client = self,
            name = name,
            local = local
        )

        return brain

    def handle_errors(self,
                      response):

        """
        Upon a bad request (400), our error log contains
        good information to raise.

        We also catch a few more common codes to
        try and print simpler messages.

        Otherwise expects this to be caught by raise_for_status()
        if applicable
        https://2.python-requests.org/en/master/_modules/requests/models/#Response.raise_for_status

        This is under the assumption that we generaly call response.json()
        after this, and that fails in poor way if there is no json available.
        """

        # Default
        if response.status_code == 200:
            return

        # Errors
        if response.status_code == 400:
            try:
                raise Exception(response.json()["log"]["error"])
            except:
                raise Exception(response.text)

        if response.status_code == 403:
            raise Exception("Invalid Permission", response.text)

        if response.status_code == 404:
            raise (Exception("404 Not Found" + response.text))

        if response.status_code == 429:
            raise Exception(
                "Rate Limited. Please add buffer between calls eg time.sleep(1). Otherwise, please try again later. Else contact us if this persists.")

        if response.status_code == 500:
            raise Exception("Internal error, please try again later.")

        raise_for_status = response.raise_for_status()
        if raise_for_status:
            Exception(raise_for_status)

    def auth(self,
             project_string_id,
             client_id = None,
             client_secret = None
             ):
        """
        Define authorization configuration

        If no client_id / secret is provided it assumes project is public
        And if project isn't public it will return a 403 permission denied.

        Arguments
            client_id, string
            client_secret, string
            project_string_id, string

        Returns
            None

        Future
            More gracefully intial setup (ie validate upon setting)
        """
        self.project_string_id = project_string_id

        if client_id and client_secret:
            self.session.auth = (client_id, client_secret)


    def set_directory_by_name(self, name):
        """

        Arguments
            self
            name, string

        """

        if name is None:
            raise Exception("No name provided.")

        # Don't refresh by default, just set from existing

        names_attempted = []
        did_set = False

        if not self.directory_list:
            self.directory_list = self.directory.get_directory_list()

        for directory in self.directory_list:

            if directory.nickname == name:
                self.set_default_directory(directory = directory)
                did_set = True
                break
            else:
                names_attempted.append(directory.nickname)

        if did_set is False:
            raise Exception(name, " does not exist. Valid names are: " +
                            str(names_attempted))


    def set_default_directory(self,
                              directory_id = None,
                              directory = None):
        """
        -> If no id is provided fetch directory list for project
        and set first directory to default.
        -> Sets the headers of self.session

        """
        if directory_id:
            self.directory_id = directory_id
        if directory:
            self.directory_id = directory.id
            self.default_directory = directory
            
        if not hasattr(self, 'directory_list'):
            self.directory_list = self.directory.get_directory_list()

        self.session.headers.update(
            {'directory_id': str(self.directory_id)})


# TODO review not using this pattern anymore

setattr(Project, "get_label_file_dict", get_label_file_dict)
setattr(Project, "convert_label", convert_label)
setattr(Project, "label_new", label_new)



from diffgram.file.file import File
from ..regular.regular import refresh_from_dict
import logging
from diffgram.core.diffgram_dataset_iterator import DiffgramDatasetIterator
from multiprocessing.pool import ThreadPool as Pool


class Directory(DiffgramDatasetIterator):

    def __init__(self, 
                 client, 
                 file_id_list_sliced = None, 
                 init_file_ids = True, 
                 validate_ids = True):

        self.client = client
        self.id = None
        self.file_list_metadata = {}
        self.nickname = None
        if file_id_list_sliced is None and init_file_ids:
            self.init_files()
        elif not init_file_ids:
            self.file_id_list = []
        elif file_id_list_sliced is not None:
            self.file_id_list = file_id_list_sliced
        super(Directory, self).__init__(self.client, self.file_id_list, validate_ids)



    def init_files(self):
        self.file_id_list = self.all_file_ids()

    def get_directory_list(self):
        """
        Get a list of available directories for a project
        """

        if self.client.project_string_id is None:
            raise Exception("No project string." + \
                            "Set a project string using .auth()")

        if type(self.client.project_string_id) != str:
            raise Exception("project_string_id must be of type String")

        endpoint = "/api/v1/project/" + self.client.project_string_id + \
                   "/directory/list"

        response = self.client.session.get(self.client.host + endpoint)

        self.client.handle_errors(response)

        data = response.json()
        
        directory_list_json = data.get('directory_list')
        default_directory_json = data.get('default_directory')

        if default_directory_json:
            self.client.directory_id = default_directory_json.get('id')

        directory_list = self.convert_json_to_sdk_object(directory_list_json)

        return directory_list
        

    def convert_json_to_sdk_object(self, directory_list_json):

        directory_list = []

        for directory_json in directory_list_json:
            new_directory = Directory(
                client = self.client,
                init_file_ids = False,
                validate_ids = False
                )
            refresh_from_dict(new_directory, directory_json)

            # note timing issue, this needs to happen after id is refreshed
            new_directory.init_files()  

            directory_list.append(new_directory)

        return directory_list



    def all_files(self):
        """
            Get all the files of the directoy.
            Warning! This can be an expensive function and take a long time.
        :return:
        """
        page_num = 1
        result = []
        while page_num is not None:
            diffgram_files = self.list_files(
                limit = 1000, 
                page_num = page_num, 
                file_view_mode = 'base')
            page_num = self.file_list_metadata['next_page']
            result = result + diffgram_files
        return result

    def all_file_ids(self, query = None):
        page_num = 1
        result = []

        diffgram_ids = self.list_files(
            limit = 5000, 
            page_num = page_num, 
            file_view_mode = 'ids_only', 
            query = query)

        if diffgram_ids is False:
            raise Exception('Error Fetching Files: Please check you are providing a valid query.')
        result = result + diffgram_ids
        page_num = self.file_list_metadata['next_page']
        total_pages = self.file_list_metadata['total_pages']
        pool = Pool(20)

        pool_results = []
        if page_num is not None:
            for i in range(page_num, total_pages + 1):
                result_async = pool.apply_async(self.list_files, (i, 5000, None, 'ids_only', None))
                pool_results.append(result_async)

        for pool_result in pool_results:
            file_ids = pool_result.get()
            result = result + file_ids

        return result

    def explore(self):
        message = '{}/studio/annotate/{}/explorer?dataset_id={}'.format(
            self.client.host,
            self.project.project_string_id,
            self.id
        )
        print('\033[92m' + 'To Explore your dataset visit:' + '\033[0m')
        print('\033[96m' + message + '\033[0m')

    def slice(self, query):
        from diffgram.core.sliced_directory import SlicedDirectory
        # Get the first page to validate syntax.
        self.list_files(
            limit = 25,
            page_num = 1,
            file_view_mode = 'ids_only',
            query = query,
        )
        sliced_dataset = SlicedDirectory(
            client = self.client,
            query = query,
            original_directory = self
        )
        return sliced_dataset

    def to_pytorch(self, transform = None):
        """
            Transforms the file list inside the dataset into a pytorch dataset.
        :return:
        """
        from diffgram.pytorch_diffgram.diffgram_pytorch_dataset import DiffgramPytorchDataset
        file_id_list = self.file_id_list
        pytorch_dataset = DiffgramPytorchDataset(
            project = self.client,
            diffgram_file_id_list = file_id_list,
            transform = transform,
            validate_ids = False

        )
        return pytorch_dataset

    def to_tensorflow(self):
        from diffgram.tensorflow_diffgram.diffgram_tensorflow_dataset import DiffgramTensorflowDataset
        file_id_list = self.file_id_list
        diffgram_tensorflow_dataset = DiffgramTensorflowDataset(
            project = self.client,
            diffgram_file_id_list = file_id_list,
            validate_ids = False
        )
        return diffgram_tensorflow_dataset

    def new(self, name: str):
        """
        Create a new directory and update directory list.

        We include name in exception message since this may
        be included in larger functions in which
        the name may be unclear

        """
        if name is None:
            raise Exception("No name provided.")

        # Confirm not in existing
        # generator expression returns True if the directory
        # is not found. this is a bit awkward.
        if next((dir for dir in self.client.directory_list
                 if dir.nickname == name), True) is not True:
            raise Exception(name, "Already exists")

        packet = {'nickname': name}

        endpoint = "/api/v1/project/" + \
                   self.client.project_string_id + "/directory/new"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = packet)

        self.client.handle_errors(response)

        data = response.json()

        project = data.get('project')
        if project:
            directory_list_json = project.get('directory_list')
            if directory_list_json:
                self.client.directory_list = self.convert_json_to_sdk_object(directory_list_json)

        new_directory = None
        # TODO the route about should return the newly created dataset directly
        for directory in self.client.directory_list:
            if directory.nickname == name:
                new_directory = directory

        return new_directory

    def list_files(
            self,
            page_num = 1,
            limit = 100,
            search_term: str = None,
            file_view_mode: str = 'annotation',
            query: str = None):
        """
        """
        if self.id:
            logging.info("Using Dataset ID " + str(self.id))
            directory_id = self.id
        else:
            logging.info("Using Default Dataset ID " + str(self.client.directory_id))
            directory_id = self.client.directory_id

        metadata = {'metadata':
            {
                'directory_id': directory_id,
                'annotations_are_machine_made_setting': "All",
                'annotation_status': "All",
                'limit': limit,
                'media_type': "All",
                'page': page_num,
                'file_view_mode': file_view_mode,
                'search_term': search_term,
                'query': query
            }
        }

        # User concept, in this context, is deprecated
        # 'sdk' is a placeholder value

        endpoint = "/api/project/" + \
                   self.client.project_string_id + \
                   "/user/sdk" + "/file/list"

        response = self.client.session.post(
            self.client.host + endpoint,
            json = metadata)

        self.client.handle_errors(response)

        # Success
        data = response.json()
        file_list_json = data.get('file_list')
        self.file_list_metadata = data.get('metadata')
        # TODO would like this to perhaps be a seperate function
        # ie part of File_Constructor perhaps
        if file_view_mode == 'ids_only':
            return file_list_json
        else:
            file_list = []
            for file_json in file_list_json:
                file = File.new(
                    client = self.client,
                    file_json = file_json)
                file_list.append(file)

            return file_list

    def get(self,
            name: str):

        """

        Returns Dataset Object for given name.
        Uses existing dir list from project.

         "NEW" version of set_directory_by_name()
         TODO refactor set_directory_by_name() to use this

        """

        if name is None:
            raise Exception("No name provided.")

        names_attempted = []
        did_set = False

        if not self.client.directory_list:
            self.client.directory_list = self.get_directory_list()

        for directory in self.client.directory_list:

            if directory.nickname == name:
                return directory

            else:
                names_attempted.append(directory.nickname)

        if did_set is False:
            raise Exception(name, " does not exist. Valid names are: " +
                            str(names_attempted))

    def add(self,
            file_list: list = None,
            file_id_list: list = None):
        """
        Add links from file_ids to the dataset.
        Does not modify existing links

        file_id_list expects ints of file_ids
        file_list is a list of Diffgram file objects
        """
        if not file_id_list and not file_list:
            raise "Must have either file_id_list or file_list (File class objects)"

        file_list_formated_as_dicts = []

        if file_id_list:
            for file_id in file_id_list:
                file_list_formated_as_dicts.append(
                    {'id': file_id})

        if file_list:
            for file in file_list:
                file_list_formated_as_dicts.append(
                    {'id': file.id})

        endpoint = "/api/v1/project/" + \
                   self.client.project_string_id + \
                   "/file/transfer"

        spec_dict = {
            'file_list': file_list_formated_as_dicts,
            'destination_directory_id': self.id,
            'mode': 'TRANSFER',
            'transfer_action': 'mirror'
        }

        response = self.client.session.post(
            self.client.host + endpoint,
            json = spec_dict)

        self.client.handle_errors(response)

        response_json = response.json()

        return response_json
